/* ez-set.c: Low level hash table.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stddef.h>
#include <assert.h>

#include "ez-set.h"

/**
 * Pretty much a traditional chained hash table implementation.  I've
 * experimented with a few variations but as a general purpose table
 * this is always competetive at worst.
 */

static void *scan_next_chain(ez_set *h, ez_set_scan *scan, int idx) __attribute__((noinline));
static void set_rehash(ez_set *h, int newsize) __attribute__((noinline));

/**
 * Lookup an entry.
 *
 * Find the entry and return the previous node in the bucket chain.
 * The previous node may be en entry in table.
 *
 * @param key key to lookup.
 * @param pp return for the parent pointer.  This will
 * be a pointer to the table entry if the chain is empty.
 * @return an entry if found.
 */
static ez_node *set_lookup(ez_set *h, const ez_node *key, unsigned int hash, ez_node **pp) {
	unsigned int idx = hash & h->mask;
	ez_node *scan = h->table[idx];
	ez_node *p = (ez_node *)&h->table[idx];

	while (scan && !(hash == scan->u.set.hash && h->equals(key, scan))) {
		p = scan;
		scan = scan->u.set.next;
	}
	*pp = p;

	return scan;
}

/**
 * Rehash the table to a new size.
 *
 * This cannot fail.  If the memory allocation fails for the table it
 * just does nothing and the set will still function mayhaps with
 * degraded performance.
 */
static void set_rehash(ez_set *h, int newsize) {
	ez_node **table = calloc(sizeof(*table), newsize);
	unsigned int mask = newsize - 1;

	if (!table)
		return;

	// Move all entries to the new table
	for (int i=0;i<=h->mask;i++) {
		ez_node *e = h->table[i];

		while (e) {
			ez_node *n = e->u.set.next;
			unsigned int nidx = e->u.set.hash & mask;

			e->u.set.next = table[nidx];
			table[nidx] = e;

			e = n;
		}
	}

	// update table
	free(h->table);
	h->table = table;
	h->mask = mask;
}

/**
 * Find the next non-empty chain.
 */
static void *scan_next_chain(ez_set *h, ez_set_scan *scan, int idx) {
	ez_node *e = NULL;

	while (e == NULL && idx <= h->mask)
		e = h->table[idx++];

	scan->e = e;
	scan->idx = idx;
	scan->p = e ? (struct ez_node *)&h->table[idx-1] : NULL;

	return e;
}

/* ********************************************************************** */
/* Main api */

void ez_set_init(ez_set *h, unsigned int (*hash)(const void *), int (*equals)(const void *, const void *), void (*efree)(void *)) {
	h->table = NULL;
	h->mask = 0;
	h->count = 0;
	h->hash = hash;
	h->equals = equals;
	h->free = efree;
}

void ez_set_clear(ez_set *h) {
	if (h->mask) {
		for (int i=0;i<=h->mask;i++) {
			ez_node *e = h->table[i];

			while (e) {
				ez_node *n = ez_node_next(e);
				if (h->free)
					h->free(e);
				e = n;
			}
			h->table[i] = NULL;
		}
		free(h->table);
		h->table = NULL;
		h->mask = 0;
		h->count = 0;
	}
}

int ez_set_size(ez_set *h) {
	return h->count;
}

void *ez_set_put(ez_set *h, void *ep) {
	ez_node *e = ep, *scan, *p;

	if (h->mask == 0) {
		h->table = calloc(sizeof(*h->table), 16);
		if (!h->table)
			return ep;
		h->mask = 15;
	}

	// Entries are only ever hashed once, here.
	e->u.set.hash = h->hash(e);
	scan = set_lookup(h, e, e->u.set.hash, &p);

	// insert new node, possibly unlinking old one
	p->u.set.next = e;
	if (scan) {
		// link old node tail
		e->u.set.next = scan->u.set.next;
	} else {
		// new node added
		e->u.set.next = NULL;
		h->count += 1;
		if (h->count > h->mask)
			set_rehash(h, (h->mask+1) * 2);
	}

	return scan;
}

void *ez_set_get(ez_set *h, const void *ep) {
	const struct ez_node *e = ep;
	struct ez_node *p;
	unsigned int hash;

	if (h->mask == 0) return NULL;

	hash = ez_set_node_hash(h, e);

	return set_lookup(h, e, hash, &p);
}

void *ez_set_remove(ez_set *h, const void *ep) {
	const ez_node *e = ep;
	struct ez_node *scan, *p;
	unsigned int hash;

	if (h->mask == 0) return NULL;

	hash = ez_set_node_hash(h, e);
	scan = set_lookup(h, e, hash, &p);
	if (scan) {
		// unlink old node
		p->u.set.next = scan->u.set.next;
		h->count -= 1;
		if (h->mask > 15 && h->count*2 < h->mask)
			set_rehash(h, (h->mask+1)>>1);
	}

	return scan;
}

/* ********************************************************************** */
/* Set iterator */

void *ez_set_scan_init(ez_set *h, ez_set_scan *scan) {
	if (h->mask == 0) return NULL;

	scan->set = h;
	scan_next_chain(h, scan, 0);

	return scan->e;
}

void *ez_set_scan_next(ez_set_scan *scan) {
	if (scan->e) {
		// advance e and p
		scan->p = scan->e;
		scan->e = scan->e->u.set.next;

		if (!scan->e)
			return scan_next_chain(scan->set, scan, scan->idx);
	}

	return scan->e;
}

void *ez_set_scan_remove(ez_set_scan *scan) {
	if (scan->e) {
		// unlink e and advance but don't change p
		scan->p->u.set.next = scan->e = scan->e->u.set.next;
		scan->set->count -= 1;

		if (!scan->e)
			return scan_next_chain(scan->set, scan, scan->idx);
	}

	return scan->e;
}

/* ********************************************************************** */
/* App helpers */

/*
 * Some hash functions.
 *
 * It's all a bit arbitrary in my experience and almost anything
 * reasonable works well enough so long as you avoid really poor ones
 * like `(int32_t)(intptr_t)pointer' or sum(char[]).
 */

// http://www.cse.yorku.ca/~oz/hash.html
//  djb2
// Use the multiplier here and let the compiler shift-ize if it that's faster.
unsigned int ez_hash_string(const char *s) {
	unsigned int hash = 5381;

	while (*s)
		//hash = (*s++) + (hash << 5) + hash;
		hash = hash * 33 + (*s++);

	return hash;
}

// http://www.cse.yorku.ca/~oz/hash.html
//  sdbm, used in berkeley db
// Use the multiplier here and let the compiler shift-ize if it that's faster.
unsigned int ez_hash_array(const void *sp, size_t size) {
	unsigned int hash = 0;
	const unsigned char *s = sp;
	const unsigned char *se = s + size;

	while (s < se)
		//hash = (*s++) + (hash << 6) + (hash << 16) - hash;
		hash = hash * 65599 + (*s++);

	return hash;
}

// https://www.burtleburtle.net/bob/hash/integer.html
//  'Thomas Wang's function'
// TBH fuck knows, an unbiased 32-bit reversible hash isn't necessarily
// great for a power of 2 hash table index.
unsigned int ez_hash_int32(uint32_t a) {
	a += ~(a<<15);
	a ^=  (a>>10);
	a +=  (a<<3);
	a ^=  (a>>6);
	a += ~(a<<11);
	a ^=  (a>>16);
	return a;
}

unsigned int ez_hash_int64(uint64_t x) {
	return ez_hash_int32(x ^ (x>>32));
}

unsigned int ez_hash_ptr(void *v) {
	if (sizeof(v) == 8)
		return ez_hash_int64((intptr_t)v);
	else
		return ez_hash_int32((intptr_t)v);
}
