/* ez-tree.c Balanced (avl) tree.

   Copyright (C) 1998-2002, 2004 Free Software Foundation, Inc.
   Copyright (C) 2019 Michael Zucchi

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 3 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>

#include "ez-tree.h"

/**
 * This is an implementation of an AVL balanced binary tree optimised
 * for code and memory size.  The source itself is also intended to be
 * concise.
 *
 * Each tree node takes only 2 pointers of overhead.  The balance
 * factor is stored in the lower 2 bits of the link pointers.  This
 * makes for some messy programming but the overhead isn't high.
 *
 * The code-size is reduced by using parameterisation and other small
 * tricks.
 *
 * As with the other ez- containers normal operation involves no
 * resource allocation.  Operations cannot fail assuming valid
 * program state.  Likewise there is almost no checking of any
 * state; the caller is entirely responsible for correct operation.
 *
 * The insert function was coded from Algorithm A in section 6.2.2 in
 * ``The Art of Computer Programming'' Volume 3, Donald E. Knuth.
 *
 * Most of the the delete function was taken from GNU libavl 2.0.  It
 * was then parameterised and broken apart into reusuable components.
 *
 * The rest I just made up.
 */

/* ********************************************************************** */

#define BALANCE0 0

/*
  Balance is stored as a 2-bit signed integer in the 2 lsb bits of
  link[0].

  A constraint is that all nodes must be allocated to at least
  4-byte boundaries, which is always the case with any known memory
  allocator.

  This is all a bit fiddly but will save at least 3 bytes per node on
  32-bit systems and 7 on 64-bit systems.
 */

static __inline__ int node_balance(struct ez_node *n) __attribute__((always_inline));
static __inline__ void node_set_balance(struct ez_node *node, int balance) __attribute__((always_inline));
static __inline__ void *node_link(struct ez_node *n, int i) __attribute__((always_inline));
static __inline__ void node_set_link(struct ez_node *n, int i, struct ez_node *link) __attribute__((always_inline));

static void *fail_depth(void) __attribute__((noinline));

static void *fail_depth(void) {
	fprintf(stderr, "EZ_TREE_MAX_DEPTH=%d exceeded\n", EZ_TREE_MAX_DEPTH);
	return NULL;
}

static __inline__ int node_balance(struct ez_node *n) {
	int i = n->u.tree.link[0];
	return i << 30 >> 30;
}

static __inline__ void node_set_balance(struct ez_node *node, int balance) {
	node->u.tree.link[0] = (node->u.tree.link[0] & ~3) | (balance & 3);
}

static __inline__ void *node_link(struct ez_node *n, int i) {
	return (void *)(n->u.tree.link[i] & ~3);
}

static __inline__ void node_set_link(struct ez_node *n, int i, struct ez_node *link) {
	n->u.tree.link[i] = (n->u.tree.link[i] & 3) | (intptr_t)link;
}

/**
 * Perform a single node rotation.
 *
 * See Knuth $6.2.3 expression (1) Case 1, where s is A, and r is B.
 * The names s and r are also from Knuth $6.2.3 algorithm A.
 *
 * @param s is really A
 * @param r is really B
 * @param ai is the link direction one takes to go from A to B.  It must be 0 or 1.
 * @param a a balance factor to set on the nodes.  This depends on
 * other state, it must be -1, 0, or 1.
 */
static struct ez_node *node_rotate_single(struct ez_node * __restrict s, struct ez_node * __restrict r, int ai, int a) {
	node_set_link(s, ai, node_link(r, ai ^ 1));
	node_set_link(r, ai ^ 1, s);

	node_set_balance(s, -a);
	node_set_balance(r, +a);

	return r;
}

/**
 * Perform a double node rotation.
 *
 * See Knuth $6.2.3 expression (1) Case 2, where s is A, and r is B.
 * The names s and r are also from Knuth $6.2.3 algorithm A.
 *
 * @param s is really A
 * @param r is really B
 * @param ai is the link direction one takes to go from A to B.  It must be 0 or 1.
 * @param a is the node comparision value matching ai.  It must be -1 or +1.
 */
static struct ez_node *node_rotate_double(struct ez_node *s, struct ez_node *r, int ai, int a) {
	struct ez_node *p = node_link(r, ai ^ 1);
	int pb = node_balance(p);

	node_set_link(r, ai ^ 1, node_link(p, ai));
	node_set_link(p, ai, r);
	node_set_link(s, ai, node_link(p, ai ^ 1));
	node_set_link(p, ai ^ 1, s);

	node_set_balance(s, pb == +a ? -a : 0);
	node_set_balance(r, pb == -a ? +a : 0);
	node_set_balance(p, 0);

	return p;
}

/**
 * Internal node remove function.
 *
 * This is much much nastier than it looks.  The rebalance code was
 * taken from libavl and paramterised etc.
 *
 * The stack is a list of all nodes navigated to @param p including
 * the node link directions taken at each level.  The stack includes
 * the psuedo-root node at stack[0] and the node to remove at sp[-1].
 *
 * @param stack stack base.
 * @param sp stack pointer to one beyond the end of the stack.
 */
static void node_remove(struct ez_tree_scan_info *stack, struct ez_tree_scan_info *sp) {
	struct ez_node *p = sp[-1].node;
	struct ez_node *r = ez_node_right(p);
	struct ez_tree_scan_info *si;

	if (!r) {
		node_set_link(sp[-2].node, sp[-2].link, ez_node_left(p));
		sp -= 1;
	} else {
		struct ez_node *s = ez_node_left(r);

		if (!s) {
			node_set_link(r, 0, ez_node_left(p));
			node_set_balance(r, node_balance(p));
			node_set_link(sp[-2].node, sp[-2].link, r);
			sp[-1] = (struct ez_tree_scan_info) { r, 1 };
		} else {
			struct ez_node *q;
			struct ez_tree_scan_info *psp = sp;

			*sp++ = (struct ez_tree_scan_info){ r, 0 };
			while ((q = ez_node_left(s))) {
				*sp++ = (struct ez_tree_scan_info){ s, 0 };
				r = s;
				s = q;
			}

			node_set_link(s, 0, ez_node_left(p));
			node_set_link(r, 0, ez_node_right(s));
			node_set_link(s, 1, ez_node_right(p));
			node_set_balance(s, node_balance(p));

			node_set_link(psp[-2].node, psp[-2].link, s);
			psp[-1] = (struct ez_tree_scan_info){ s, 1 };
		}
	}

	// Rebalance tree
	// basic code from libavl but paramterised as in Knuth
	si = sp-1;
	while (si > stack) {
		struct ez_node *y = si->node;
		int ai = si->link;
		int aj = ai ^ 1;
		int a = ai * 2 - 1;
		int yb = node_balance(y);

		if (a == yb)
			node_set_balance(y, 0);
		else if (yb == 0) {
			node_set_balance(y, -a);
			break;
		} else {
			struct ez_node *x = node_link(y, aj);
			int xb = node_balance(x);
			struct ez_node *w;

			if (xb == a)
				w = node_rotate_double(y, x, aj, -a);
			else
				w = node_rotate_single(y, x, ai ^ 1, xb == 0 ? a : 0);

			node_set_link(si[-1].node, si[-1].link, w);

			if (xb == 0)
				break;
		}

		si -= 1;
	}

	stack[0].node->u.tree.link[0] -= 1;
}

/**
 * Recursive tree free operation.
 *
 * Doesn't update any nodes.
 */
static void tree_free(ez_tree *tree, struct ez_node *n, ez_free_fn node_free) {
	while (n) {
		struct ez_node *l = ez_node_left(n);
		struct ez_node *r = ez_node_right(n);

		tree_free(tree, r, node_free);
		node_free(n);
		n = l;
	}
}

/* ********************************************************************** */

void ez_tree_init(struct ez_tree *tree, ez_cmp_fn node_cmp) {
	tree->root.u.tree.link[0] = 0;
	tree->root.u.tree.link[1] = 0;
	tree->node_cmp = node_cmp;
}

void ez_tree_clear(ez_tree *tree, ez_free_fn node_free) {
	tree_free(tree, ez_tree_root(tree), node_free);
	tree->root.u.tree.link[0] = 0;
	tree->root.u.tree.link[1] = 0;
}

void *ez_tree_get(ez_tree *tree, const void *node) {
	const struct ez_node *k = node;
	struct ez_node *p = ez_tree_root(tree);
	ez_cmp_fn node_cmp = tree->node_cmp;

	while (p) {
		int cmp = node_cmp(k, p);

		if (cmp)
			p = node_link(p, cmp > 0);
		else
			break;
	}

	return p;
}

void *ez_tree_put(ez_tree *tree, void *node) {
	struct ez_node *k = node;
	struct ez_tree_scan_info stack[EZ_TREE_MAX_DEPTH];
	struct ez_tree_scan_info *sp = stack, *se = sp + EZ_TREE_MAX_DEPTH - 1;
	ez_cmp_fn node_cmp = tree->node_cmp;

	k->u.tree.link[0] = BALANCE0; // balance=0
	node_set_link(k, 1, 0);

	if (!tree->root.u.tree.link[1]) {
		tree->root.u.tree.link[1] = (intptr_t)k;
		tree->root.u.tree.link[0] += 1;
		return NULL;
	}

	// Search for node, keeping track of last non-balanced node in si
	struct ez_tree_scan_info *si = stack+1;
	int cmp = 1;
	struct ez_node *p, *n;

	*sp++ = (struct ez_tree_scan_info){ &tree->root, 1 };
	n = p = ez_node_right(&tree->root);
	while (n && cmp != 0) {
		if (sp >= se)
			return fail_depth();
		p = n;
		cmp = node_cmp(k, p);
		*sp++ = (struct ez_tree_scan_info){ p, cmp > 0 };
		n = node_link(p, cmp > 0);
		if (n && node_balance(n))
			si = sp;
	}

	// Matched, replace and return old node
	if (cmp == 0) {
		node_set_link(sp[-2].node, sp[-2].link, k);
		*k = *p;
		return p;
	}

	// Insert new node
	node_set_link(p, cmp > 0, k);
	tree->root.u.tree.link[0] += 1;
	*sp = (struct ez_tree_scan_info){ k, 0 };

	// Fix balance factors between s and q
	for (struct ez_tree_scan_info *pi = si+1;pi != sp; pi+=1)
		node_set_balance(pi->node, pi->link * 2 - 1);

 	// Balance
	struct ez_node *s = si[0].node;
	struct ez_node *r = si[1].node;
	int ai = si->link;
	int a = ai * 2 - 1; // map 0,1 to -1,+1
	int balance = node_balance(s);

	if (balance == 0) {
		node_set_balance(s, a);
	} else if (balance == -a) {
		node_set_balance(s, 0);
	} else {
		struct ez_node *q;

		if (node_balance(r) == a)
			q = node_rotate_single(s, r, ai, 0);
		else
			q = node_rotate_double(s, r, ai, a);

		node_set_link(si[-1].node, si[-1].link, q);
	}

	return NULL;
}

void *ez_tree_remove(ez_tree *tree, const void *key) {
	struct ez_node *p = ez_node_right(&tree->root);
	struct ez_node *n = p;
	struct ez_tree_scan_info stack[EZ_TREE_MAX_DEPTH];
	struct ez_tree_scan_info *sp = stack, *se = sp + EZ_TREE_MAX_DEPTH;
	ez_cmp_fn node_cmp = tree->node_cmp;
	int cmp = 1;

	*sp++ = (struct ez_tree_scan_info) { &tree->root, 1 };
	while (n && cmp != 0) {
		if (sp >= se)
			return fail_depth();
		p = n;
		cmp = node_cmp(key, p);
		*sp++ = (struct ez_tree_scan_info) { p, cmp > 0 };
		n = node_link(p, cmp > 0);
	}

	if (cmp == 0)
		node_remove(stack, sp);

	return p;
}

static void *ez_tree_scan_dir(ez_tree_scan *scan, enum ez_node_link_t dir) {
	int l = scan->level;
	int nir = dir ^ 1;
	struct ez_node *p = scan->stack[l-1].node;
	struct ez_node *r = node_link(p, dir);

	if (r) {
		scan->stack[l-1].link = dir;
		do {
			if (l >= EZ_TREE_MAX_DEPTH)
				return fail_depth();
			scan->stack[l++] = (struct ez_tree_scan_info) { r, nir };
			r = node_link(r, nir);
		} while (r);
		scan->level = l;
		return scan->stack[l-1].node;
	} else {
		while (l > 2 && scan->stack[l-2].link == dir)
			l--;
		scan->level = l - 1;
		if (l > 2)
			return scan->stack[l-2].node;
		return NULL;
	}
}

void *ez_tree_scan_init_key(ez_tree *tree, ez_tree_scan *scan, enum ez_node_link_t scan_dir, enum ez_node_link_t key_dir, void *key) {
	struct ez_node *p = ez_node_right(&tree->root);
	ez_cmp_fn node_cmp = tree->node_cmp;
	int l = 0;
	int cmp = 1;

	scan->tree = tree;
	scan->scan_dir = scan_dir;
	scan->stack[l++] = (struct ez_tree_scan_info) { &tree->root, 1 };

	if (key) {
		while (p && cmp != 0) {
			if (l >= EZ_TREE_MAX_DEPTH)
				return fail_depth();
			cmp = node_cmp(key, p);
			scan->stack[l++] = (struct ez_tree_scan_info) { p, cmp > 0 };
			p = node_link(p, cmp > 0);
		}
	} else {
		while (p) {
			if (l >= EZ_TREE_MAX_DEPTH)
				return fail_depth();
			scan->stack[l++] = (struct ez_tree_scan_info) { p, 1 - scan_dir };
			p = ez_node_link(p, 1 - scan_dir);
		}
	}
	scan->level = l;

	if (l > 1) {
		if (key) {
			int c = (cmp > 0) - (cmp < 0); // -1 0 1 from compare
			int a = key_dir * 2 - 1;       // -1 x 1 from direction

			if (c == a)
				return ez_tree_scan_dir(scan, key_dir);
			// same as:
			//if ((cmp < 0 && dir == EZ_LINK_LEFT)
			//    || (cmp > 0 && dir == EZ_LINK_RIGHT))
			//	return ez_tree_scan_next(scan, dir);
		}

		return scan->stack[l-1].node;
	} else
		return NULL;
}

void *ez_tree_scan_init(ez_tree *tree, ez_tree_scan *scan, enum ez_node_link_t scan_dir) {
	return ez_tree_scan_init_key(tree, scan, scan_dir, 1 - scan_dir, NULL);
}

void *ez_tree_scan_next(ez_tree_scan *scan) {
	return ez_tree_scan_dir(scan, scan->scan_dir);
}

void *ez_tree_scan_prev(ez_tree_scan *scan) {
	return ez_tree_scan_dir(scan, 1 - scan->scan_dir);
}

void *ez_tree_scan_remove(ez_tree_scan *scan, enum ez_node_link_t dir) {
	int l = scan->level;
	struct ez_tree_scan_info *sp = scan->stack + l;

	if (l > 1) {
		struct ez_node *last = sp[-1].node;

		node_remove(scan->stack, sp);

		// If node_remove could maintain the stack or indicate when it
		// is still valid this could be avoided, but its messy.

		if (dir != -1)
			return ez_tree_scan_init_key(scan->tree, scan, scan->scan_dir, dir, last);
	}
	return NULL;
}

void *ez_tree_scan_put(ez_tree_scan *scan, void *node) {
	struct ez_node *n = node;
	struct ez_tree_scan_info *sp = scan->stack + scan->level;
	struct ez_node *o = sp[-1].node;

	*n = *o;
	sp[-1].node = n;
	node_set_link(sp[-2].node, sp[-2].link, n);

	return o;
}
